//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// Parameter
#include <XSModel/Parameter/Parameter.h>
// SumComponent
#include <XSModel/Model/Component/SumComponent.h>
// Expression
#include <XSUtil/Parse/Expression.h>
// Component
#include <XSModel/Model/Component/Component.h>
// ComponentCreator
#include <XSModel/Model/Component/ComponentCreator.h>
// CGroupBase
#include <XSModel/Model/Component/CGroupBase.h>

#include <XSModel/Model/UniqueEnergy.h>
#include <XSModel/Model/Component/AddComponent.h>
#include <XSModel/Model/Component/AddTableComponent.h>
#include <XSModel/Model/Component/MulComponent.h>
#include <XSModel/Model/Component/MulTableComponent.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSUtil/Parse/XSparse.h>
#include <XSUtil/Utils/XSstream.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSstreams.h>
#include <XSsymbol.h>
#include <XSModel/Model/ModelBase.h>
#include <iterator>
#include <typeinfo>
#include <memory>


// Class CGroupBase 
int CGroupBase::s_count = 0;

CGroupBase::CGroupBase(const CGroupBase &right)
  : m_index(right.m_index), 
    m_nested(right.m_nested),
    m_groupFlux(right.m_groupFlux->clone(this)),  // owning
    m_componentInfo(right.m_componentInfo),// non-owning. 
    m_parent(right.m_parent),
    m_elements()
{
//  m_groupFlux = right.m_groupFlux->clone(this); 
  ComponentListConstIterator ic = right.m_elements.begin();
  ComponentListConstIterator icEnd = right.m_elements.end();
  while ( ic != icEnd)
  {
        m_elements.push_back((*ic)->clone(this));   
        ++ic;    
  }     
  ++s_count;
}

CGroupBase::CGroupBase (Expression& gs, ModelBase* p, bool nest)
  : m_index(++s_count), 
    m_nested(nest),
    m_groupFlux(new SumComponent(this)),  // owning
    m_componentInfo(&gs),// non-owning.
    m_parent(p),  
    m_elements()
{
}


CGroupBase::~CGroupBase()
{
  delete m_groupFlux;

  ComponentListIterator it = m_elements.begin();
  ComponentListIterator itEnd = m_elements.end();
  while (it != itEnd)
  {
      //SumComponent's are owned by their ComponentGroup - don't delete them here
      if(!dynamic_cast<SumComponent*>(*it))
	  delete *it;

      ++it;
  }

  --s_count;
}


CGroupBase & CGroupBase::operator=(const CGroupBase &right)
{
  CGroupBase temp(right);
  swap(temp);
  return *this;
}


void CGroupBase::swap (CGroupBase& right)
{
  std::swap(m_index,right.m_index);
  std::swap(m_nested,right.m_nested);
  std::swap(m_groupFlux,right.m_groupFlux);
  std::swap(m_componentInfo,right.m_componentInfo);
  std::swap(m_parent,right.m_parent);
  std::swap(m_elements,right.m_elements);
}

void CGroupBase::resetComponentFlux ()
{
  ComponentListConstIterator end=m_elements.end();
  ComponentListConstIterator beg=m_elements.begin();

  for (ComponentListConstIterator it = beg ;  it!=end; ++it)
  {
        (*it)->restoreUniquePhotonArray();
  }
}

std::vector<ModParam*> CGroupBase::getVariableParameters () const
{
  ComponentListConstIterator cl = m_elements.begin();
  ComponentListConstIterator clEnd = m_elements.end();

  std::vector<ModParam*> vp;
  while (cl != clEnd)
  {
        std::vector<ModParam*> cvp = (*cl)->getVariableParameters();
        std::copy(cvp.begin(),cvp.end(),back_inserter(vp));
        ++cl;       
  }
  return vp;
}

SumComponent* CGroupBase::subgroupSum (ComponentListIterator begin, ComponentListIterator end)
{

  ComponentListIterator add = begin;
  // the following circumlocution is a workaround for deficiencies in the
  // SunPRo 6.0 compiler. It doesn't recognize the std::distance function
  // on list iterators.
  int n (0);
  while (begin++ != end) ++n;
  begin = add;
//  int n = std::distance(copy.begin(),copy.end());

  // construct a SumComponent from it
  SumComponent* accum = 0;
  if ( n > 1 )
  {
        // find the pointer to AddComponent.
        AddComponent* gen(0);
        while ( add != end && !(gen = dynamic_cast<AddComponent*>(*add))) ++add;
        if ( add != end )
        {
              // apply the normalization parameter to the additive model,
              // except where the grand summation happens. The nested evaluation
              // algorithm is designed to flatten all nested subcomputations
              // to that level.
              accum = new SumComponent(*gen);
              accum->normalize();
        }
        ComponentListIterator postAdd = add;
        if (postAdd != end)  ++postAdd;
        while (postAdd != end)
        {
                try
                {
                        MulComponent& postMultiplier = dynamic_cast<MulComponent&>(**postAdd);
                        // the non-virtual SumComponent::operator*=(MulComponent&)
                        *accum *= postMultiplier;
                        ++postAdd;
                }
                catch (std::bad_cast)
                {
                        throw RedAlert("Syntax: postmultiplier is not multiplicative");       
                }   
        }
        ComponentListIterator pre (add);
        while ( pre != begin )                 
        {
                // the virtual Component::operator*(const SumComponent&)
                --pre;       
                *accum = (**pre) * (*accum);
        }        
  }
  else
  {
        AddComponent* gen = dynamic_cast<AddComponent*>(*add);
        if ( !gen )
        {
              accum = static_cast<SumComponent*>(*add);         
        } 
        else 
        {
              accum = new SumComponent(*gen);
              accum->normalize();
        }
  }
  return accum;            
}

const string& CGroupBase::parentName () const
{

  return m_parent->name();
}

CGroupBase* CGroupBase::clone () const
{

  return new CGroupBase(*this);
}

void CGroupBase::setElement (XSModExpTree<ComponentGroup*>& componentGroups)
{
   int gt=0;
   int N (m_componentInfo->nWords());
   for ( int j = 0; j < N; ++j)
   {
      ComponentCreator create;
      const string& componentName = m_componentInfo->words(j);
      gt = componentName.find(ComponentGroup::GROUPTEST());
   // if this is not a symbol placeholder for a nested group...
   //  create a component of the appropriate type and read it.
      if ( gt < 0  )
      {                
         m_elements.push_back(create.GetComponent(componentName,this));
         incrementModelComponentCount();
         Component* newComponent = m_elements.back();
	 newComponent->index(m_componentInfo->wordNumber(j));
         updateExpression(newComponent, componentName, j);
      }
      else
   // if it is, store the address of the SumComponent that's in the
   // location nn given by  $GROUPnn.
      {
         int w = 0;
         string groupNo = componentName.substr(componentName.find_first_not_of(ComponentGroup::GROUPTEST()));
         sscanf (groupNo.c_str(),"%d",&w);
         ComponentGroup* cg = componentGroups.values()[w];
         SumComponent* gf = cg->groupFlux();

         gf->name(componentName);
         m_elements.push_back(gf);
      }
   }
}

size_t CGroupBase::dataGroup () const
{

  return m_parent->dataGroupNumber();
}

void CGroupBase::registerParameters () const
{
  ComponentListConstIterator rE(m_elements.begin());
  ComponentListConstIterator rEnd(m_elements.end());

  while (rE != rEnd)
  {
        (*rE)->registerParameters();
        ++rE;       
  }
}

size_t CGroupBase::parameterIndexBase () const
{

  return m_parent->parameterIndexBase();
}

size_t CGroupBase::incrementModelParameterCount ()
{

  return m_parent->incrementModelParameterCount();
}

size_t CGroupBase::incrementModelComponentCount ()
{

  return m_parent->incrementModelComponentCount();
}

size_t CGroupBase::modelIndex () const
{

  return m_parent->index();
}

Parameter* CGroupBase::getLocalParameter (size_t i) const
{

  return m_parent->getLocalParameter(i);
}

Parameter* CGroupBase::localParameter (size_t i) const
{
  ComponentListConstIterator rc(m_elements.begin());      
  ComponentListConstIterator rcEnd(m_elements.end());      
  Parameter* result(0);     
  while (rc != rcEnd)
  {
        if ( (result = (*rc)->localParameter(i)) != 0 ) break;
        ++rc;
  }
  return  result;
}

void CGroupBase::setElements (const string& componentName, const Component* newElement)
{
  ComponentListIterator rs(m_elements.begin());
  ComponentListIterator rsEnd(m_elements.end());


  while ( rs != rsEnd )
  {
        if  (  (*rs)->name() == componentName ) 
        {
                rs = m_elements.erase(rs);
                m_elements.insert(rs,newElement->clone(this));
                break;
        }
        ++rs;       
  }
}

void CGroupBase::calculate (bool saveComponentFlux, bool frozen)
{
  ComponentListIterator end=m_elements.end();
  ComponentListIterator beg=m_elements.begin();

  // scheme: perform the model calculations if needed.
  for (ComponentListIterator it = beg ;  it!=end; ++it)
  {
        // recompute component if parameters have changed.
        // the parameter changed attribute sets the recompute flag
        // to true if one or more of the parameters have changed.
        if ((*it)->recompute() | (*it)->isGroup() | frozen ) 
        {
                (*it)->calculate(saveComponentFlux,frozen);  
                // most of the time, we've calculated the model and don't need
                // to change it again until the parameter is changed. In the
                // case of 'frozen' used in computing norms, we will need to recompute
                // the entire model.
        }
 }
}

const ArrayContainer& CGroupBase::energyArray () const
{

  return m_parent->energy();
}

SumComponent* CGroupBase::parenSum (ComponentListIterator& first, ComponentListIterator& last)
{

        // any output from here is at chatter level 20 or higher
        XSstream::verbose(tpout,25,25);
        //tcout << "CGroupBase::parenSum\n";



        // declare an auto_ptr for the result of combining the components in
        // parentheses
        std::auto_ptr<SumComponent> totalGroupFlux;

        int jwords(0);
        int j(0);
        int lb ( m_componentInfo->lbrace().size() ? m_componentInfo->lbrace(0) : 0);
        while (m_componentInfo->wordLocation(jwords) < lb ) ++jwords;
        ComponentListIterator begin = first;

        while (begin != last  )
        {
                ComponentListIterator  run = begin;
                if ( j < m_componentInfo->nPlus())
                {
                        int pl = m_componentInfo->plus(j);        
                        // 1) find an iterator to the first word after lb.
                        // 2) find an iterator to the last word before pl.
                        // create the group sum.  
                        while (m_componentInfo->wordLocation(jwords) < pl) 
                        {
                                ++jwords;
                                ++run;   
                        } 
                        if ( j == 0)  
                        {
                                SumComponent* term(subgroupSum(begin,run));
                                if ( !term->isGroup() ) 
                                {
                                        totalGroupFlux.reset(term);  
                                }
                                else
                                {
                                        totalGroupFlux.reset(term->clone(this));
                                }
                                totalGroupFlux->name("SUM");    
                        }
                        else
                        {
                                SumComponent* term(subgroupSum(begin,run));
                                *totalGroupFlux += *term;
                                if ( !term->isGroup() ) delete term;
                        }
                        ++j;                    
                        begin = run;     
                }
                else
                {
                        if ( false )
                        {

                                tcout << "\nBegin " << (*run)->name() << '\n';                       
                                tcout << "\nEnd " << " (end iterator)" << '\n';                   
                                tcout << std::flush;                
                        }
                        SumComponent* term(subgroupSum(run,last));
                        *totalGroupFlux += *term;
                        if ( !term->isGroup() ) delete term;
                        begin = last;

                }
        }


        // output arrays if in high verbose mode
        totalGroupFlux->debugPrint(tcout,"\nModel Output CGroupBase::parenSum()");


        XSstream::verbose(tpout);
        return totalGroupFlux.release();
}

void CGroupBase::setComputeFlag (bool value)
{
  ComponentListConstIterator end=m_elements.end();
  ComponentListConstIterator beg=m_elements.begin();

  // scheme: perform the model calculations if needed.
  // recompute component if parameters have changed.
  // the parameter changed attribute sets the recompute flag
  // to true if one or more of the parameters have changed.
  // on initialization this might switch computing off altogether.
  // during a fit one probably needs manual control over this flag
  // as well. In principle also value could be false to use only
  // the last computed model.
  for (ComponentListConstIterator it = beg ;  it!=end; ++it)  (*it)->recompute(value);
}

std::list<ModParam*> CGroupBase::normParams () const
{
  std::list<ModParam*> norms;
  ComponentListConstIterator cEnd = m_elements.end();
  for (ComponentListConstIterator c = m_elements.begin(); c != cEnd; ++c)
  {
                  // ignore the SumComponents that represent groups.
          if ( !(*c)->isGroup())
          {
                AddComponent* a = dynamic_cast<AddComponent*> (*c);
                if (a)
                {
                        // every AddComponent has exactly one norm
                        // parameter of type ModParam, added on construction.
                        ModParam* p = static_cast<ModParam*>(a->parameter("norm"));
                        norms.push_back(p);
                }
          }            
  }       

  return norms;
}

void CGroupBase::clearArrays (const std::set<UniqueEnergy*>& currentEngs)
{
  m_groupFlux->clearArrays(currentEngs);

  ComponentListIterator cEnd = m_elements.end();
  for (ComponentListIterator c = m_elements.begin(); c != cEnd; ++c)
  {                        
          if ( !(*c)->isGroup()) (*c)->clearArrays(currentEngs);
  }       
}

void CGroupBase::checkNorms (IntegerArray& paramsToFreeze) const
{

  size_t zeroNorms (0);
  std::list<ModParam*> norms = normParams();
  std::list<ModParam*>::iterator ni = norms.begin();
  std::list<ModParam*>::iterator niEnd = norms.begin();

  // find out if there are any zero norms.

  while ( ni != niEnd )
  {
          if ( (*ni)->value('v') < SMALL ) ++zeroNorms;
  }
  if ( zeroNorms != 0 )
  {
          // if there are. Then, if all the norms are zero, add all of the
          // relevant param numbers to the 'to be temporarily frozen' list.

          if (zeroNorms == norms.size())
          {
                  ComponentListConstIterator cEnd = m_elements.end();
                  for (ComponentListConstIterator c = m_elements.begin(); c != cEnd; ++c)
                  {                        
                          std::vector<ModParam*> vp = (*c)->getVariableParameters();
                          size_t N = vp.size();
                          for ( size_t j = 0; j < N; ++j)
                          {
                                  paramsToFreeze.push_back(parent()->index() << 16 + 
                                                  vp[j]->index());
                          }
                  }
          }
          else
          {
                  // if however only some of the norms are frozen, then add
                  // the parameters of the 
                  ComponentListConstIterator cEnd = m_elements.end();
                  for (ComponentListConstIterator c = m_elements.begin(); c != cEnd; ++c)
                  {         
                          AddComponent* ac = dynamic_cast<AddComponent*>(*c);
                          if ( ac )    
                          {           
                                  std::vector<ModParam*> vp = ac->getVariableParameters();

                                  size_t N (vp.size());
                                  bool zero (false);
                                  for ( size_t j = 0; j < N; ++j)
                                  {
                                          if ( vp[j]->name() == "norm")
                                          {
                                                  if (vp[j]->value() < SMALL ) 
                                                  { 
                                                          zero = true;
                                                          break;
                                                  }       
                                          }
                                  }
                                  if ( zero )
                                  {
                                          for ( size_t j = 0; j < N; ++j)
                                          {
                                                  paramsToFreeze.push_back(parent()->index() 
                                                                  << 16 +  vp[j]->index());      
                                          }

                                  }       
                          }
                  }

          }
  }       
}

void CGroupBase::saveComponentFlux (bool setSaveFlux)
{
  ComponentListConstIterator end=m_elements.end();
  ComponentListConstIterator beg=m_elements.begin();

  for (ComponentListConstIterator it = beg ;  it!=end; ++it)
  {
        (*it)->saveUniquePhotonArray(setSaveFlux);
  }
}

void CGroupBase::combine ()
{
  ComponentListIterator end=m_elements.end();
  ComponentListIterator beg=m_elements.begin();
  ComponentListIterator it=m_elements.begin();
  // combine nested subgroups. This is a recursive call to this function.
  while (it != end)
  {
     Component* current(*it);
     if (current->isGroup()) current->combine(); 
     ++it;
  }

  // sum the components.
  bool noParen(m_componentInfo->nLbrace() == 0);

  if (noParen)
  {
     SumComponent* tmp(subgroupSum(beg,end));
     *m_groupFlux = *tmp;
     if (!tmp->isGroup()) delete tmp;       
  }
  else
  {                
     ComponentListIterator parBeg(m_elements.begin());
     // there will be exactly one set of parentheses because of the definition
     // of a component group and the treatment of nested groups.  
     // if there are no '+' signs in the parenthesis the job is a lot easier,
     // so treat this as a special case.

     bool multipleSubgroups( m_componentInfo->nPlus() > 0 );

     int jwords(0);

     // after model editing it is possible that there are no braces.
     // in which case sum everything by setting left and right brace
     // delimiters to the beginning and end of the string.

     int lb(m_componentInfo->lbrace().size() ? m_componentInfo->lbrace(0) : 0 );
     int rb(m_componentInfo->rbrace().size() ? m_componentInfo->rbrace(0) 
                     : m_componentInfo->parseString().size() );         
             // 
     // start point   
     //

     while (m_componentInfo->wordLocation(jwords) < lb) 
     {
        ++parBeg;
        ++jwords;
     }

     ComponentListIterator parEnd(parBeg);
     while (parEnd != m_elements.end() && m_componentInfo->wordLocation(jwords) < rb)
     {
        ++jwords;
        ++parEnd;
     }

     // replace the parentheses by a single SumComponent in a copy 
     // of the Component List and sum the result.

     std::auto_ptr<SumComponent> parenFlux(0);       
     if (!multipleSubgroups)
     {
        parenFlux.reset(subgroupSum(parBeg,parEnd));       
     }
     else
     {
        parenFlux.reset(parenSum(parBeg,parEnd));
     } 
     ComponentList tmp(m_elements); 
     ComponentListIterator tmpBeg(tmp.begin());
     ComponentListIterator it(m_elements.begin());
     while ( it != parBeg )
     {
        ++it;
        ++tmpBeg;
     }
     ComponentListIterator tmpEnd(tmpBeg);
     while (it !=parEnd)
     {
        ++it;
        ++tmpEnd;
     }
     tmp.insert(tmpBeg, parenFlux.get());
     tmp.erase(tmpBeg, tmpEnd);
     SumComponent* tmpSum = finalSum(tmp.begin(),tmp.end());
     *m_groupFlux = *tmpSum;
     // Due to above assignment, m_groupFlux has picked up tmpSum's 
     // Parameter* to an AddComp norm parameter.  m_groupFlux should
     // not retain this, otherwise if this is a nested CGroup, the
     // enclosing CGroup's SumComponent placeholder pointer will also
     // point to this norm parameter.  This condition caused parameter
     // linking errors when a data group copy model was made.
     std::vector<Parameter*> empty;
     m_groupFlux->setParameterSet(empty);
     if (!tmpSum->isGroup()) delete tmpSum;

  }

  // output arrays if in high verbose mode


  m_groupFlux->debugPrint(tcout," Model Output CGroupBase::combine() -");

}

std::list<SumComponent*> CGroupBase::sources ()
{
   std::list<SumComponent*> sourceList;
   ComponentListIterator beg = m_elements.begin();
   ComponentListIterator end = m_elements.end();
   bool noParen(m_componentInfo->nLbrace()==0);
   if (noParen)
   {
      // This component group is of the form:
      // [mult comp]*..*[conv comp]*ADDCOMP*[mult comp]...
      // subGroupSum WILL allocate memory for a new SumComponent.
      SumComponent *sumComp = subgroupSum(beg, end);
      sourceList.push_back(sumComp);
      return sourceList;
   }
   else
   {
      // There are 1 and only 1 set of parentheses.
      // Find the first and last of the elements inside 
      // the parentheses. (parEnd will point to the element AFTER the 
      // right parenthesis, or to the end.)
      ComponentListIterator parBeg = beg;
      int iwords = 0, jwords = 0;
      int lb(m_componentInfo->lbrace().size() ? m_componentInfo->lbrace(0) : 0 );
      int rb(m_componentInfo->rbrace().size() ? m_componentInfo->rbrace(0) 
                      : m_componentInfo->parseString().size() );         
      while (m_componentInfo->wordLocation(iwords) < lb) 
      {
              ++parBeg;
              ++iwords;
      }
      jwords = iwords;
      ComponentListIterator parEnd = parBeg;
      while (parEnd != end && m_componentInfo->wordLocation(jwords) < rb)
      {
              ++jwords;
              ++parEnd;
      }
      // The parenList function will recursively call back on this
      // function for any nested groups between the parentheses.
      sourceList = parenList(iwords, parBeg, parEnd);
      // Now apply any mult or conv components that appear outside
      // the parentheses.  First, apply the post components.
      while (parEnd != end)
      {
         try
         {
            MulComponent& postMultiplier = dynamic_cast<MulComponent&>(**parEnd);
            std::list<SumComponent*>::iterator itSl = sourceList.begin();
            std::list<SumComponent*>::iterator itSlEnd = sourceList.end();
            while (itSl != itSlEnd)
            {
               **itSl *= postMultiplier;
               ++itSl;
            } 
         }          
         catch (std::bad_cast)
         {
            throw RedAlert("Syntax: postmultiplier is not multiplicative");
         }
         ++parEnd;
      }
      // Now apply the pre components.
      while (parBeg != beg)
      {
         --parBeg;
         std::list<SumComponent*>::iterator itSl = sourceList.begin();
         std::list<SumComponent*>::iterator itSlEnd = sourceList.end();
         while (itSl != itSlEnd)
         {
            **itSl = (**parBeg) * (**itSl);
            ++itSl;
         }          
      }
   }

   return sourceList;     
}

std::list<SumComponent*> CGroupBase::parenList (int iWord, const ComponentListIterator& first, const ComponentListIterator& last)
{
  std::list<SumComponent*> sourceList;
  ComponentListIterator startSubGroup = first;  
  int iPlus=0;      
  while (iPlus <= m_componentInfo->nPlus())
  {
     SumComponent *groupTest = dynamic_cast<SumComponent*>(*startSubGroup);
     if (groupTest)
     {
        // For nested groups, don't do anything on this level,
        // but handle it with a RECURSIVE CALL to CGroupBase::sources().
        std::list<SumComponent*> groupList = groupTest->sources();
        sourceList.insert(sourceList.end(), groupList.begin(), groupList.end());

        // A group must be the ONLY word and component in 
        // between plus signs.
        ++iPlus;
        ++iWord; 
        ++startSubGroup;
     }
     else
     {
        ComponentListIterator endSubGroup;
        if (iPlus == m_componentInfo->nPlus())
        {
           endSubGroup = last;
        }
        else
        {
           int pl = m_componentInfo->plus(iPlus);
           endSubGroup = startSubGroup;
           // If this else block is reached, the while loop below
           // should have to execute at least once....Another way
           // of saying that if there is still a plus sign to process,
           // we must be starting with a word (and component) to the
           // left of it.
           while (m_componentInfo->wordLocation(iWord) < pl)
           {
              ++iWord;
              ++endSubGroup;
           }           
        }
        // subgroupSum will allocate the sumComp memory.
        SumComponent *sumComp = subgroupSum(startSubGroup, endSubGroup);
        sourceList.push_back(sumComp);        
        startSubGroup = endSubGroup;
        ++iPlus;
     }
  }

  return sourceList;
}

SumComponent* CGroupBase::finalSum (ComponentListIterator begin, ComponentListIterator end)
{

  ComponentListIterator sum = begin;
  // the following circumlocution is a workaround for deficiencies in the
  // SunPRo 6.0 compiler. It doesn't recognize the std::distance function
  // on list iterators.
  int n (0);
  while (begin++ != end) ++n;
  begin = sum;
//  int n = std::distance(copy.begin(),copy.end());

  // gen (generator) a (possibly new) SumComponent object used to accumulate
  // into. If it represents a nested group, it's a temporary anyway.
  SumComponent* gen (0);
  SumComponent* accum(0);
  if ( n > 1)
  {
        // find the pointer to the SumComponent.
        while ( sum != end && !(gen = dynamic_cast<SumComponent*>(*sum))) ++sum;
        // Save this location for doing pre-factors later.
        ComponentListIterator pre = sum;
        if ( !gen->isGroup())
        {
             accum = gen->clone(this);   
        }
        else
        {
             accum = gen;       
        }
        ++sum;
        while (sum != end)
        {
                try
                {
                        MulComponent& postMultiplier  = dynamic_cast<MulComponent&>(**sum);
                        // the non-virtual SumComponent::operator*=(MulComponent&)
                        *accum *= postMultiplier;
                        ++sum;  
                }
                catch (std::bad_cast)
                {
                        throw RedAlert("Syntax: postmultiplier is not multiplicative");       
                }   
        }
        while ( pre != begin )                 
        {
                // the virtual Component::operator*(const SumComponent&)
                --pre;       
                *accum = (**pre) * (*accum);
        }        
  }
  // construct a SumComponent from it
  else
  {
        gen = dynamic_cast<SumComponent*>(*sum);          
        if (!gen->isGroup())
        {
                accum = gen->clone(this);
        }
        else
        {
                accum = gen;       
        }
  }


  return accum;   
}

size_t CGroupBase::modelComponentCount ()
{

  return m_parent->modelComponentCount();
}

size_t CGroupBase::modelParameterCount ()
{

  return m_parent->modelParameterCount();
}

void CGroupBase::deleteComponent (int sequenceNumber, int index)
{
  ComponentListIterator doomed(m_elements.begin());
  for (int ic = 0; ic < sequenceNumber; ++ic) ++doomed;
  delete *doomed;
  m_elements.erase(doomed);
  m_componentInfo->deleteWordbySequence(index);
  m_componentInfo->update();
}

void CGroupBase::setNestedElement (const string& componentName, Component* newElement)
{
  ComponentListIterator rs(m_elements.begin());
  ComponentListIterator rsEnd(m_elements.end());

  // nested element does not reset pointer to current component group.
  while ( rs != rsEnd )
  {
        if  (  (*rs)->name() == componentName ) 
        {
                // delete current pointer and replace with nested version.
                // setNestedElement is called from ModelBase::clone which has
                // just created ptrs like rs.
                delete *rs;
                rs = m_elements.erase(rs);
                m_elements.insert(rs,newElement);
                break;
        }
        ++rs;       
  }  
}

void CGroupBase::resetElement (XSModExpTree<ComponentGroup*>& newGroups, const std::vector<Component*>& existingComponents)
{
   const size_t nComps = static_cast<size_t>(m_componentInfo->nWords());
   for (size_t i=0; i<nComps; ++i)
   {
      const string& componentName = m_componentInfo->words(i);
      if (componentName.find(ComponentGroup::GROUPTEST()) ==
            string::npos)
      {
         // This gets the word sequence, which corresponds to component 
         // index number.
         int newIndex = m_componentInfo->wordNumber(i);
         Component* compToTransfer = existingComponents[newIndex-1];
         compToTransfer->index(newIndex);
         compToTransfer->setParent(this);
         m_elements.push_back(compToTransfer);
      }
      else
      {
         // Store the address of the SumComponent that's in
         // the location nn given by  $GROUPnn.
	 int w = 0;
	 string groupNo = componentName.substr(componentName.find_first_not_of(ComponentGroup::GROUPTEST()));
	 sscanf (groupNo.c_str(),"%d",&w);
         ComponentGroup* cg = newGroups.values()[w];
	 SumComponent* gf = cg->groupFlux();

	 gf->name(componentName);
	 m_elements.push_back(gf);
      }
   }
}

Component* CGroupBase::element (int index) const
{
  Component* target(0);
  ComponentListConstIterator cl = m_elements.begin();
  ComponentListConstIterator clEnd = m_elements.end();
  while ( cl != clEnd && !target )
  {
        if ( (*cl)->index() == index ) target = *cl;
        ++cl;       
  }

  return target;
}

void CGroupBase::insertComponent (int location, int index, const string& name, char op)
{
    ComponentListIterator ic (m_elements.begin());
    ComponentCreator create;
    Component::getSetupData(name);
    string fullName(Component::currentModelName());
    Component* newComponent (create.GetComponent(fullName,this));
    newComponent->index(index);
    m_componentInfo->insertWordbySequence(index,fullName,op);

    addToElements(location, newComponent);

    ic = m_elements.begin();
    ComponentListIterator icEnd = m_elements.end();
    int i(0);
    // Ensure word sequence for ComponentList and Expression match,
    // leaving out the groups if any.
    while (ic != icEnd )
    {
	if(!(*ic)->isGroup()) 
	    m_componentInfo->setWordNumber(i,(*ic)->index());
	++ic, ++i;
    }
    // This call is only needed in case of table models, particularly
    // if a re-prompt for file name was necessary.
    updateExpression(newComponent, fullName, location);
    incrementModelComponentCount();
}

void CGroupBase::exchangeComponent (int location, Component* newComponent)
{
   // Introduced for use by editmod exchange:
   // Simply replace an existing component pointer with a new one
   // created from the input name.  This will DESTROY the old
   // component, so it ASSUMES all parameter deregistering has
   // been done already.
   Component* oldComponent = 0;
   int count = 0;
   ComponentListIterator itComp = m_elements.begin();
   ComponentListIterator itEnd = m_elements.end();
   if (location < m_elements.size())
   {
      while (count != location && itComp != itEnd)
      {      
         ++itComp;
         ++count;
      }
      if (itComp != itEnd)
         oldComponent = *itComp;
   }
   if (!oldComponent)
   {
      std::ostringstream oss;
      oss << "Error searching for component at location " << location
         << " in CGroupBase::exchangeComponent.\n";
      throw RedAlert(oss.str());
   }
   *itComp = newComponent;
   newComponent->index(oldComponent->index());
   delete oldComponent;
}

Component* CGroupBase::createComponent (string& name)
{
   // Introduced for use in editmod, to ensure new component can be created 
   // before modifying anything in its ultimately intended model.  This
   // function can throw.
   ComponentCreator create;
   Component::getSetupData(name);
   string fullName(Component::currentModelName());
   Component* newComponent (create.GetComponent(fullName,this));
   // file name may have changed due to re-prompting while in
   // create.GetComponent.
   string tableModName = checkCurrentTableModelExpr(newComponent, name);
   if (tableModName.length())
      name = tableModName;
   return newComponent;
}

void CGroupBase::deregisterParameters () const
{
  ComponentListConstIterator rE(m_elements.begin());
  ComponentListConstIterator rEnd(m_elements.end());

  while (rE != rEnd)
  {
        (*rE)->deregisterParameters();
        ++rE;       
  }
}

Component* CGroupBase::firstComponent ()
{
  if ( !m_elements.empty() ) return m_elements.front();
  else return 0;
}

void CGroupBase::resizeElements (size_t n)
{
    m_elements.resize(n);
}

void CGroupBase::addToElements (int componentOffset, Component* newComponent)
{
    ComponentListIterator ic = m_elements.begin();

    if ( componentOffset < static_cast<int>(m_elements.size()))
    {
        for (int j = 0; j < componentOffset; ++j)
	    ++ic;

        ic = m_elements.insert(ic, newComponent);
    }
    else // no indexes to update.
	m_elements.push_back(newComponent);
    newComponent->setParent(this);
}

const std::set<UniqueEnergy*>& CGroupBase::uniqueEnergyArray () const
{
   return m_parent->getUniqueEnergies();
}

void CGroupBase::updateExpression (Component* newComponent, const string& oldName, size_t wordIdx)
{
   // If this is a table model, it's possible the file name has 
   // changed as a result of re-prompting.  Therefore we need
   // to re-insert this possibly new name back into the expression.
   string tableName(checkCurrentTableModelExpr(newComponent, oldName));
   if (tableName.length())
      m_componentInfo->setWord(tableName, wordIdx);
   else
      m_componentInfo->setWord(newComponent->name(),wordIdx);
   m_componentInfo->update();
}

string CGroupBase::checkCurrentTableModelExpr (const Component* component, const string& prevExprString)
{
   string newComponentName;
   if (const AddTableComponent* addTable = dynamic_cast<const AddTableComponent*>(component))
   {
      const string& newFileName = addTable->tableFile();
      newComponentName = "atable{" + newFileName + "}";
   }
   else if (const MulTableComponent* mulTable = dynamic_cast<const MulTableComponent*>(component))
   {
      const string& newFileName = mulTable->tableFile();
      // This could begin with mtable or etable.  If it made it
      // this far, oldName has already been validated and only
      // need to check 'eE' or 'mM'.
      if (prevExprString[0] == 'e' || prevExprString[0] == 'E')
         newComponentName = "etable{";
      else
         newComponentName = "mtable{";
      newComponentName += newFileName + "}";
   }
   return newComponentName;
}

// Additional Declarations
