//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// SumComponent
#include <XSModel/Model/Component/SumComponent.h>
// ComponentBase
#include <XSModel/Model/Component/ComponentBase.h>
// ConComponent
#include <XSModel/Model/Component/ConComponent.h>

#include <XSFunctions/XSModelFunction.h>
#include <XSModel/Model/UniqueEnergy.h>
#include <typeinfo>


// Class ConComponent 

ConComponent::ConComponent(const ConComponent &right)
   : Component(right)
{
  cloneParameters(*this,right);
}

ConComponent::ConComponent (CGroupBase* p)
        : Component(p)
{
}


ConComponent::~ConComponent()
{
  componentBase()->deleteParameters();
}


SumComponent& ConComponent::operator *= (SumComponent& right) const
{
  const std::set<UniqueEnergy*>& energies = getUniqueEnergies();
  std::set<UniqueEnergy*>::const_iterator en   = energies.begin();
  std::set<UniqueEnergy*>::const_iterator enEnd = energies.end();

  const std::vector<Parameter *>&  parameterSet = componentBase()->parameterSet();
  size_t  numParams(parameterSet.size());
  RealArray  startValues(numParams);
  for (size_t  i=0; i<numParams; i++)
  {
     startValues[i] = parameterSet[i]->value();
  }

  while (en != enEnd)
  {
     const UniqueEnergy* uniqueEng = *en;      
     if (right.uniquePhotonArray(uniqueEng).size()) 
     {
        RealArray convErr;
        RealArray& flux = right.uniquePhotonArray(uniqueEng);
        size_t spectrumNumber = *uniqueEng->clientSpectra().begin();

        const XSModelFunction*  generator(componentBase()->generator());
        (*generator)(uniqueEng->energy(), startValues, spectrumNumber, flux, 
                        convErr,componentBase()->initString());

        // warning: no convolution model has errors to propagate as yet. But, this
        // looks like the method to use. If there is error in the additive model
        // then fluxErrorArray has finite size. The error output from the generator
        // should be the error in the convolution convolved with the fluxArray. 
        if ( right.isError() || isError())
        {
          RealArray& rErr = right.uniquePhotonErrArray(uniqueEng);
          if (right.isError())
          {
             RealArray convolvedFluxErr(sqrt(rErr));
             RealArray secondOrderError;  // dummy.
             (*generator)(uniqueEng->energy(), startValues, spectrumNumber, convolvedFluxErr, 
                           secondOrderError, componentBase()->initString());
             rErr = convolvedFluxErr*convolvedFluxErr;
          }

        }
     } 
     ++en;
  }
  return right;
}

ConComponent* ConComponent::clone (CGroupBase* p) const
{
  ConComponent* cloned = new ConComponent(*this);
  cloned->setParent(p);
  return cloned;    
}

void ConComponent::copy (const Component& right)
{
  if ( typeid(*this) != typeid(right) ) 
          throw RedAlert("*** Programmer Error: copying to wrong type ***");  
  ConComponent __temp(static_cast<const ConComponent&>(right));
  swap(__temp);   
}

SumComponent ConComponent::operator * (const SumComponent& right) const
{
  SumComponent __tmp(right);
  return (*this) *= __tmp;  
}

void ConComponent::calculate (bool saveComponentFlux, bool frozen)
{
  if (saveComponentFlux) saveUniquePhotonArray(saveComponentFlux);


  // do nothing else - computations done in *= operator.      
}

// Additional Declarations
