//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// ComponentBase
#include <XSModel/Model/Component/ComponentBase.h>
// MixComponent
#include <XSModel/Model/Component/MixComponent.h>
#include <XSstreams.h>
#include <XSFunctions/XSModelFunction.h>
#include <XSModel/Model/Component/MixBase.h>
#include <XSModel/Model/Component/CGroupBase.h>
#include <typeinfo>
#include <memory>


// Class MixComponent 

MixComponent::MixComponent(const MixComponent &right)
  : Component(right)
{
  cloneParameters(*this,right);
}

MixComponent::MixComponent (CGroupBase* p)
     : Component(p)
{
}


MixComponent::~MixComponent()
{
  componentBase()->deleteParameters(); 
}


MixComponent* MixComponent::clone (CGroupBase* p) const
{
  MixComponent* cloned = new MixComponent(*this);
  cloned->setParent(p);
  return cloned;    
}

void MixComponent::copy (const Component& right)
{
  if ( typeid(*this) != typeid(right) ) 
          throw RedAlert("*** Programmer Error: copying to wrong type ***");  
  MixComponent __temp(static_cast<const MixComponent&>(right));
  swap(__temp);    
}

void MixComponent::calculate (bool saveComponentFlux, bool frozen)
{
  if (saveComponentFlux) saveUniquePhotonArray(saveComponentFlux);


  // do nothing else. calculate at the 'multiply' stage.      
}

void MixComponent::initialize ()
{
  // initialize method to be called once per model container update.  
  if ( componentBase()->mixGenerator() )
  {
          const std::vector<Parameter*>& parameterSet = componentBase()->parameterSet();
          size_t  numParams = parameterSet.size();
          std::vector<Real>  startValues(numParams);
          for (size_t j = 0; j < numParams; ++j)
          {
                  startValues[j] = parameterSet[j]->value();
          }

          const string& modelName = parent()->parentName();
          componentBase()->mixGenerator()->initialize(startValues, modelName);
  }
}

void MixComponent::perform (const EnergyPointer& energy, GroupFluxContainer& flux, GroupFluxContainer& fluxError)
{

  // this perform method must pass information from the SumComponent to be massaged
  // by the implementation of the mixing model component.

  const XSModelFunction*  generator(componentBase()->generator());

  // parameterValues
  const std::vector<Parameter*>& parameterSet = componentBase()->parameterSet();
  size_t  numParams = parameterSet.size();
  std::vector<Real>  startValues(numParams);
  for (size_t j = 0; j < numParams; ++j)
  {
          startValues[j] = parameterSet[j]->value();
  }

  // The MixGenerator pointer is of type MixBase and passes an instance
  // of a model specific class implementation to the generator function.
  // this ensures persistence and memory management. The MixBase* object
  // must be constructed on first call and
  // will be deleted by the ComponentBase dtor.
  // If the model is not implemented in C++
  // the generator can ignore this ptr address.  But needs a different way
  // of ensuring persistence and memory management if those are appropriate.
  MixBase* mixGenPtr = componentBase()->mixGenerator();
  bool unInitialized (!mixGenPtr);
  std::auto_ptr<MixBase*>  pMixGeneratorHandle (new MixBase*(mixGenPtr));  
  MixBase** mixGeneratorHandle (pMixGeneratorHandle.get());
  const string& modelName = parent()->parentName();
  (*generator)(energy, startValues, flux, fluxError, mixGeneratorHandle, modelName);
  if ( unInitialized && *mixGeneratorHandle ) componentBase()->mixGenerator(*mixGeneratorHandle);
}

void MixComponent::verifyData () const
{
  if ( componentBase()->mixGenerator() ) componentBase()->mixGenerator()->verifyData();
}

void MixComponent::initializeForFit ()
{
  // called once per fit and only for first data group.
  if (root()->dataGroupNumber() == 1)
  {
     if ( componentBase()->mixGenerator() )
     {
             const std::vector<Parameter*>& parameterSet = componentBase()->parameterSet();
             size_t  numParams = parameterSet.size();
             bool parametersAreFrozen = true;
             std::vector<Real>  startValues(numParams);
             for (size_t j = 0; j < numParams; ++j)
             {
                     startValues[j] = parameterSet[j]->value();
                     if (!parameterSet[j]->isFrozen())
                     {
                        parametersAreFrozen = false;
                     }
             }


             componentBase()->mixGenerator()->initializeForFit(startValues, 
                                   parametersAreFrozen);
     }
  }
}

// Additional Declarations
