#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <locale>
#include <unistd.h>
#include <cstdlib>

using std::string;
using std::vector;
vector<size_t> processModelFile(std::istream& initFile, 
                                std::ostream& headerFile, 
                                std::ostream& codeFile);

string lowerCase(const string& inputString);


string expandDirectoryPath (const string& input);

const string funcMapContainer = "XSFunctionMap";
const string f77linkPrefix = "";
const string f77linkSuffix = "_";
const string f77funcType("xsf77Call");
const string F77funcType("xsF77Call");
const string CXXfuncType("XSCCall");
const string ccfuncType("xsccCall");
const string ccmixfuncType("xsmixcall");
const string CXXmixfuncType("XSMixCCall");

vector<size_t> processModelFile(std::istream& initFile, 
                                std::ostream& headerFile, 
                                std::ostream& codeFile,
                                const string& packageName)
{
        using namespace std;
        vector<size_t> numbersOfComponents(5,0);
        
        vector<string> headerFileLines;
        vector<string> codeFileLines;
        
        vector<string> componentTypes;
        componentTypes.push_back("add");
        componentTypes.push_back("mul");
        componentTypes.push_back("con");
        componentTypes.push_back("mix");
        componentTypes.push_back("acn");        
        
        string line ("");
        while ( getline(initFile,line) )
        {
	       int jType(-1), count = 0;
	       bool found(false);
	       string token, delim = " \t\n";
	       string::size_type end = 0, start = line.find_first_not_of(delim);
	    
	       while(start != string::npos && count < 6)
	       {
		   if((end = line.find_first_of(delim, start)) == string::npos)
		       end = line.size();
		   else if((end - start) == 1 && line[start] == '\"')
		       end = (end = line.find('\"', end)) == string::npos ? line.size() : end + 1;
		   
		   token = line.substr(start, end - start);

		   start = line.find_first_not_of(delim, end);
		   
		   ++count;
	       }

	       if(count == 6)
	       {
		   for (size_t j = 0; j < componentTypes.size(); ++j)
		   {
		       if(token == componentTypes[j])
		       {
			   found = true;
			   jType = j;
			   break;
		       }
		   }
	       }               
               
               if ( found )
               {
                        std::ostringstream hf;
                        std::ostringstream cf;

                        string modelName("");
                        int npar(0);
                        double enlow(0);
                        double enhigh(0);
                        string modType("");
                        string function(""); 
                        string funcType("");
                        istringstream modLine(line);   
                        modLine >> modelName >> npar >> enlow >> enhigh >> function >> modType;
                        string linkerFunction("");
                        if ( modType == "mix")
                        {
                                if (function.substr(0,2) == "C_")
                                {
                                        linkerFunction = function.substr(2);
                                        funcType = CXXmixfuncType;       
                                }
                                else if ( function.substr(0,2) == "c_")
                                {
                                        linkerFunction = function.substr(2);
                                        funcType = ccmixfuncType;       
                                }
                        }
                        else
                        {
                                if ( function.substr(0,2) == "C_")
                                {
                                        // C++ call
                                        linkerFunction = function.substr(2);  
                                        funcType = CXXfuncType;     
                                }
                                else if ( function.substr(0,2) == "c_")
                                {
                                        linkerFunction = function.substr(2);  
                                        funcType = ccfuncType;     
                                }
                                else if ( function.substr(0,2) == "F_")
                                {
                                   // double precision f77 call
                                   // all fortran calls need to be in lowercase (at least
                                   // on Solaris and Linux).
                                   string lowFn(function.substr(2));
                                   for (size_t j=0; j<lowFn.size(); ++j)
                                   {
                                      lowFn[j] = std::tolower(lowFn[j]);
                                   }
                                   linkerFunction = f77linkPrefix + lowFn + f77linkSuffix;
                                   funcType = F77funcType;
                                }
                                else
                                {
                                        // f77call
                                        string lowFn(function.size(),' ');
                                        for (size_t j = 0; j < function.size(); ++j) 
                                        {
                                                lowFn[j] = std::tolower(function[j]);
                                        }
                                        linkerFunction = f77linkPrefix + lowFn + f77linkSuffix;
                                        funcType = f77funcType;
                                }
                        }        
                        hf << "\t" << left << setw(10) << funcType << "   "  
                                        << linkerFunction << ";\n" ;
                        cf << "\t" << funcMapContainer << "[\""  << modelName << "\"]" 
                                 << setw(8-modelName.size() + 3) << " = " 
                                 << "new XSCall<" << funcType 
                                 << ">(" << linkerFunction << ");\n" ;
                        
                        if (linkerFunction == packageName)
                        {
                                cerr << "Error: package name cannot be identical to " 
                                     << " the name of a model function (" << linkerFunction
                                     <<  ")... exiting\n";
                                abort();    
                                               
                        }
                        else if (function.substr(2).find_first_of("_$") != string::npos)
                        {
                                cerr << "Error: model function name must not contain "
                                     << "\"_\" or \"$\" (" << linkerFunction << ") ... exiting\n";
                                abort();
                        }
                        
                        
                        headerFileLines.push_back(hf.str());
                        codeFileLines.push_back(cf.str());
                        
                        cout.setf(ios::left);
                        cout    << "Model name: " << setw(8) << modelName 
                                << "  No. of params: " << setw(4) << npar 
                                << "  Emin: " 
                                << setiosflags(ios::scientific)
                                << setw(12) << setprecision(2) 
                                << setiosflags(ios::right) << enlow 
                                << "  Emax: " 
                                << setw(12) << setiosflags(ios::right) 
                                << setprecision(2) << enhigh 
                                << "  f(): " << setw(8) << linkerFunction
                                << "  Type: " << modType << endl;

			numbersOfComponents[jType]++;
                }
        }
        size_t Nmodels (headerFileLines.size());
        for (size_t  j = 0; j < Nmodels; ++j)
        {
                headerFile << headerFileLines[j];
                codeFile << codeFileLines[j];       
        }
        return numbersOfComponents;       
}


string expandDirectoryPath (const string& input)
{
   using std::cerr;
   string stringDir("");
   char* enValue(0);
   size_t slash(input.find_first_of('/'));
   string root(input.substr(0,slash));
   switch ( input[0] )
   {
      // No need to check for "~" as in XSPEC's version of this function.
      // Any "~" should already have been expanded by the shell before
      // being passed into initpackage's main function.
      case '.':
         if (input.length() > 1 && input[1] != '.' && slash != 1)
         {
            cerr << "*** invalid directory path: must begin with absolute path"
                  << "\n*** valid environment variable or shell character "
                  << " {~,./,../}\n";
            return string("");     

         }
         else
         {
            const size_t PATHSZ = 1024;
            if ( (enValue = getcwd(NULL,PATHSZ)) == NULL)
            {
               cerr << "*** XSUtil error while converting to absolute path name."
                   << "\n*** Current dir path name may possibly exceed max size = "
                   << PATHSZ << std::endl; 
               return string("");
            }
            if (input.length() == 1)
            {
               // case of just "."
               stringDir = string(enValue);
            }
            else
            {
               switch (input[1])
               {
                  case '/':
                     // input is "./[...]"
                     stringDir = string(enValue);
                     if (slash != string::npos)
                     {
                        stringDir += input.substr(slash);       
                     }
                     break;
                  case '.':
                     {
                        // input is "..[...]"
                        stringDir = string(enValue);
                        // stringDir is an abs path so 
                        // there has to be a slash.
                        size_t lastSlash 
                                = stringDir.find_last_of('/');
                        stringDir = stringDir.substr(0,lastSlash);
                        // this looks right...
                        if (slash != string::npos)
                        {
                           stringDir 
                              += input.substr(slash);   
                        }      
                     }
                     break;
                  default:
                     break;       
               }
            } // end input.length() > 1
            free(enValue);       
         }  // end case '.'
         break;
      case '$':
         enValue = getenv(root.substr(1).c_str());
         if ( enValue ) 
         {        
             stringDir = string(enValue);
             if (slash != string::npos)
             {
                 stringDir += input.substr(slash);
             }
         }
         else
         {
            cerr << " Invalid environment variable " 
                            << root.substr(1) << '\n';
            return string("");     
         }
         break;
      case '/':
      default:
         stringDir = input;
         break;
   }
   return stringDir;        
}


string lowerCase(const string& inputString)
{
     size_t n(inputString.length());
     string outputString(inputString);
     for (size_t l = 0 ; l < n; ++l)
     {
              outputString[l] = tolower(inputString[l]);          
     }
     // this looks extremely dumb but is a workaround for a g++ compiler
     // bug.
     return outputString.substr(0,n);  
}
