//
//  XSPEC12  November 2003
//
//

#include <XSUser/Global/XSGlobal.h>  
#include <XSUser/Global/Global.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/StatMethod.h>
#include <XSFit/Fit/FitMethod.h>
#include <XSModel/Data/DataUtility.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSstreams.h>
#include <XSContainer.h>

int
XSGlobal::xsStatistic(ClientData cdata,Tcl_Interp* tclInterp,int objc, Tcl_Obj* CONST  objv[])
{
   using namespace XSContainer;
   string cmd = string(static_cast<char*>(cdata));
   string arg;
   if (objc == 1 || (arg = Tcl_GetString(objv[1])) == "?")
   {
       printDocs(cmd.c_str(),"?");
       tcout <<"(Currently using: " 
             << fit->statMethod()->fullName() <<")" <<std::endl;
   } 
   else
   {
      try
      {
         StatMethod* newStatMethod = StatMethod::get(arg);
         if ( newStatMethod ) 
         {
            if (newStatMethod->fullName() == fit->statMethod()->fullName())
            {
               tcout <<"Already using statistic: "
                  << fit->statMethod()->fullName() << std::endl;
            }
            else
            {
               const string oldWeight = DataUtility::statWeight().name();
               newStatMethod->setWeight(globalData->weight());
               // The new stat might not have accepted the global weight 
               // setting, so need to check DataUtility again to see
               // what's actually being used.
               const string newWeight = DataUtility::statWeight().name();
               if (newWeight != oldWeight)
               {
                  // Need to recalculate all variances.  Fit update 
                  // will be called later when statMethod is reset.
                  // Note for future: This calculation and xsWeight's
                  // are identical and should be placed in a new 
                  // function somewhere.
                  size_t nSpec = datasets->numberOfSpectra();
                  for (size_t i=0; i<nSpec; ++i)
                  {
                     SpectralData* sd = datasets->lookup(i+1);
                     if (sd)
                     {
                        RealArray norm = sd->areaScale()*sd->exposureTime();
                        if (Background* bkg = sd->getBackground())
                        {
                           RealArray bkgNorm = bkg->data()->areaScale()*bkg->data()->exposureTime();
                           bkgNorm *= bkg->data()->backgroundScale()/sd->backgroundScale();
                           DataUtility::statWeight()(*(bkg->data()), bkgNorm);
                        }
                        DataUtility::statWeight()(*sd, norm);
                     }
                  }
               }               
               fit->statMethod(newStatMethod);
            }
         }   
         else throw Fit::NoSuchStatMethod(arg);
      }
      catch ( YellowAlert& )
      {
          printDocs(cmd.c_str(),"?");
          return globalData->autoSave(TCL_ERROR);   
      }
   }
   tcout << std::flush;
   return globalData->autoSave(TCL_OK);
}






